from multidim_indexing.view import MultidimView, classproperty
import numpy as np


class NumpyMultidimView(MultidimView):
    @classproperty
    def lib(cls):
        return np

    @property
    def int(self):
        return int

    def arr(self, *args, **kwargs):
        return np.array(*args, **kwargs)

    def zeros(self, *args, **kwargs):
        return np.zeros(*args, **kwargs)

    @classmethod
    def cast(cls, arr, dtype):
        return arr.astype(dtype)

    @classmethod
    def all(cls, arr, dim=0):
        return np.all(arr, axis=dim)

    @classmethod
    def is_valid_arr_value(cls, val, valid):
        return not np.isscalar(val) and val.size == valid.size

    @classmethod
    def ravel_multi_index(cls, key, shape):
        if len(key.shape) == 2:
            key = key.T
        elif len(key.shape) == 3:
            key = key.transpose((1, 2, 0))
        return np.ravel_multi_index(tuple(key), shape)

    def unravel_key(self, key):
        return np.stack(np.unravel_index(key, self.shape)).T

    @classmethod
    def transpose(cls, arr):
        return np.transpose(arr)

    @classmethod
    def repeat(cls, arr, repeats):
        return np.tile(arr, repeats)

    @classmethod
    def cat(cls, arrs, dim=0):
        return np.concatenate(arrs, axis=dim)

    @classmethod
    def stack(cls, arrs, dim=0):
        return np.stack(arrs, axis=dim)



import torch
from typing import List, Tuple, Union
from multidim_indexing.view import MultidimView, classproperty


class TorchMultidimView(MultidimView):
    def __init__(self, source, *args, **kwargs):
        self.device = source.device
        super(TorchMultidimView, self).__init__(source, *args, **kwargs)

    @classproperty
    def lib(cls):
        return torch

    @property
    def int(self):
        return torch.long

    def arr(self, *args, **kwargs):
        return torch.tensor(*args, device=self.device, **kwargs)

    def zeros(self, *args, **kwargs):
        return torch.zeros(*args, device=self.device, **kwargs)

    @classmethod
    def cast(cls, arr, dtype):
        return arr.to(dtype=dtype)

    @classmethod
    def all(cls, arr, dim=0):
        return torch.all(arr, dim=dim)

    @classmethod
    def is_valid_arr_value(cls, val, valid):
        return torch.is_tensor(val) and torch.numel(val) == torch.numel(valid)

    @classmethod
    def ravel_multi_index(cls, key, shape):
        return ravel_multi_index(key, shape)

    def unravel_key(self, key):
        return unravel_index(key, self.shape)

    @classmethod
    def transpose(cls, arr):
        return arr.transpose(0, 1)

    @classmethod
    def repeat(cls, arr, repeats):
        return arr.repeat(*repeats)

    @classmethod
    def cat(cls, arrs, dim=0):
        return torch.cat(arrs, dim=dim)

    @classmethod
    def stack(cls, arrs, dim=0):
        return torch.stack(arrs, dim=dim)

# filling in functions from numpy from francois-rozet
Shape = Union[List[int], Tuple[int, ...], torch.Size]


def ravel_multi_index(coords: torch.Tensor, shape: Shape) -> torch.Tensor:
    r"""Converts a tensor of coordinate vectors into a tensor of flat indices.
    This is a `torch` implementation of `numpy.ravel_multi_index`.
    Args:
        coords: A tensor of coordinate vectors, (*, D).
        shape: The source shape.
    Returns:
        The raveled indices, (*,).
    """

    shape = coords.new_tensor(shape + (1,))
    coefs = shape[1:].flipud().cumprod(dim=0).flipud()

    return (coords * coefs).sum(dim=-1)


def unravel_index(indices: torch.Tensor, shape: Shape) -> torch.Tensor:
    r"""Converts a tensor of flat indices into a tensor of coordinate vectors.
    This is a `torch` implementation of `numpy.unravel_index`.
    Args:
        indices: A tensor of flat indices, (*,).
        shape: The target shape.
    Returns:
        The unraveled coordinates, (*, D).
    """

    shape = indices.new_tensor(shape + (1,))
    coefs = shape[1:].flipud().cumprod(dim=0).flipud()

    return torch.div(indices[..., None], coefs, rounding_mode='trunc') % shape[:-1]



import time
import numpy as np
from multidim_indexing import numpy_view as view


def test_index_2d():
    high = 50
    N = 15
    shape = (5, 10)
    data = np.arange(0, high).reshape(shape)
    data_view = view.NumpyMultidimView(data)
    key_ravelled = np.random.randint(0, high=high, size=(N,))
    key = data_view.unravel_key(key_ravelled)

    query = data_view[key]
    # since the values are just a range, the ravelled key is the queried value
    assert np.allclose(key_ravelled, query)

    # invalidate last key
    key[-1][0] = high
    query = data_view[key]
    # check that the last key is an invalid value
    assert query[-1] == data_view.invalid_value
    assert np.allclose(key_ravelled[:-1], query[:-1])


def test_index_multi_d():
    high = 500
    N = 15
    shape = (5, 10, 2, 5)
    data = np.arange(0, high).reshape(shape)
    data_view = view.NumpyMultidimView(data)
    key_ravelled = np.random.randint(0, high=high, size=(N,))
    key = data_view.unravel_key(key_ravelled)

    query = data_view[key]
    # since the values are just a range, the ravelled key is the queried value
    assert np.allclose(key_ravelled, query)


def test_batch_key_value_query():
    high = 50
    B = 3
    N = 20
    shape = (5, 10)
    data = np.arange(0, high).reshape(shape)
    data_view = view.NumpyMultidimView(data.astype(dtype=float), value_ranges=[(0, 1), (0, 5)])

    key_ravelled = np.random.randint(0, high=high, size=(N,))
    key = data_view.unravel_key(key_ravelled)
    batch_key = np.tile(key, (B, 1, 1))
    value_key = data_view.ensure_value_key(key)
    # value_key = torch.randn((N, len(shape)), dtype=torch.float)
    batch_value_key = np.tile(value_key, (B, 1, 1))
    assert not data_view.is_key_ravelled(key)
    assert not data_view.is_key_ravelled(batch_key)
    # test different ways of generating batch keys are equivalent
    batch_value_key2 = data_view.ensure_value_key(batch_key)
    assert np.allclose(batch_value_key, batch_value_key2)
    batch_key2 = data_view.ensure_index_key(batch_value_key)
    assert np.allclose(batch_key, batch_key2)

    query = data_view[value_key]
    # since the values are just a range, the ravelled key is the queried value
    assert np.allclose(key_ravelled.astype(dtype=float), query)

    # test the usage of a batch key
    batch_query = data_view[batch_value_key]
    assert batch_query.shape == (B, N)
    for i in range(B):
        assert np.allclose(batch_query[i], query)

    batch_value_key3 = np.random.randn(B, N, len(shape))
    batch_query = data_view[batch_value_key3]
    for i in range(B):
        query = data_view[batch_value_key3[i]]
        assert np.allclose(batch_query[i], query)


def test_value_2d():
    high = 50
    N = 15
    shape = (5, 10)
    data = np.arange(0, high).reshape(shape)
    data_view = view.NumpyMultidimView(data, value_ranges=[(0, 1), (0, 1)])
    key_ravelled = np.random.randint(0, high=high, size=(N,))
    index_key = data_view.unravel_key(key_ravelled)
    # convert to value between 0 and 1
    key = index_key / (np.array(shape) - 1)

    query = data_view[key]
    # since the values are just a range, the ravelled key is the queried value
    assert np.allclose(key_ravelled, query)

    # invalidate last key
    key[-1][0] = high
    query = data_view[key]
    # check that the last key is an invalid value
    assert query[-1] == data_view.invalid_value
    assert np.allclose(key_ravelled[:-1], query[:-1])


def test_set():
    high = 50
    N = 15
    shape = (5, 10)
    data = np.arange(0, high).reshape(shape)
    data_view = view.NumpyMultidimView(data)
    # having repeating indices results in undefined behavior
    key_ravelled = np.random.permutation(high)[:N]
    key = data_view.unravel_key(key_ravelled)

    data_view[key] = -5
    # test that we changed the original data
    for i in range(N):
        assert int(data[tuple(key[i])]) == -5

    rand_val = np.random.randint(-50, high=-5, size=(N,))
    data_view[key] = rand_val

    for i in range(N):
        assert data[tuple(key[i])] == rand_val[i]

    # invalidate last key
    key[-1][0] = high
    data_view[key] = rand_val[-1] + 1
    # check that we didn't change the location of the invalid key
    assert data.reshape(-1)[key_ravelled[-1]] == rand_val[-1]


def test_performance():
    B = 256
    dx = 2
    N = 512
    shape = (B, 64, 64)
    high = np.prod(np.array(shape)).astype(dtype=int)
    data = np.arange(0, high).reshape(shape)

    runs = 10

    e_ours = []
    e_builtin = []
    e_builtin2 = []
    e_shorthand = []

    for run in range(runs):
        key_ravelled = np.random.randint(0, high=high, size=(B, N,))
        key = np.stack(np.unravel_index(key_ravelled, shape), axis=0)
        for i in range(B):
            key[0, i, :] = i
        key_ravelled = np.ravel_multi_index(key, shape)
        key = key.transpose((1, 2, 0))

        key_builtin = key[:, :, 1:]
        start = time.time()
        nb = np.arange(B).reshape(-1, 1)
        idxs = np.array_split(key_builtin, key_builtin.shape[-1], axis=-1)
        idxs = [ix.squeeze(-1) for ix in idxs]
        query_builtin = data[nb, idxs[0], idxs[1]]
        elapsed_builtin = time.time() - start
        e_builtin.append(elapsed_builtin)

        start = time.time()
        query_builtin2 = data[nb, key_builtin[:, :, 0], key_builtin[:, :, 1]]
        e_builtin2.append(time.time() - start)

        start = time.time()
        data_view = view.NumpyMultidimView(data, check_safety=False)
        query = data_view[key]
        elapsed_ours = time.time() - start
        e_ours.append(elapsed_ours)

        start = time.time()
        # check that we can get the default behavior of batch indexing by default being applied to that batch
        # e.g. B x N x 2 indexing into B x X x Y
        data_view = view.NumpyMultidimView(data, check_safety=False)
        query_shorthand = data_view[key_builtin]
        e_shorthand.append(time.time() - start)

        # since the values are just a range, the ravelled key is the queried value
        assert np.allclose(key_ravelled, query)
        assert np.allclose(key_ravelled, query_shorthand)
        assert np.allclose(key_ravelled, query_builtin)
        assert np.allclose(key_ravelled, query_builtin2)

    scale = 1000
    e_ours = np.array(e_ours) * scale
    e_shorthand = np.array(e_shorthand) * scale
    e_builtin = np.array(e_builtin) * scale
    e_builtin2 = np.array(e_builtin2) * scale
    print(f"elapsed ours {np.mean(e_ours)} ({np.std(e_ours)}) "
          f"ours shorthand {np.mean(e_shorthand)} ({np.std(e_shorthand)}) "
          f"builtin {np.mean(e_builtin)} ({np.std(e_builtin)}) "
          f"builtin2 {np.mean(e_builtin2)} ({np.std(e_builtin2)})")


if __name__ == "__main__":
    test_index_2d()
    test_index_multi_d()
    test_set()
    test_value_2d()
    test_performance()



import abc


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class MultidimView(abc.ABC):
    def __init__(self, source, value_ranges=None, invalid_value=-1, check_safety=True):
        """
        View into a tensor or numpy array that is convenient to index using a batch of indices.
        Intended for use on a cache of a function that needs to be indexed into with coordinates
        rather than integer indices.

        :param source: d dimensional underlying tensor or numpy array
        :param value_ranges: d pairs of [min, max] coordinate keys for each of the d dimensions. It is an inclusive
            range, so indexing with the boundary values of either min or max is valid
        :param invalid_value: when check_safety is True, querying with out of range coordinates returns this value
            rather than raising an error. Needs to be the same dtype as source
        :param check_safety: whether to check if the keys are within bound; turn off to get about 40% speedup in
            indexing
        """
        self.dtype = source.dtype
        self.shape = source.shape
        self.dim = len(source.shape)
        self.invalid_value = invalid_value
        self.check_safety = check_safety

        if value_ranges is not None:
            self._min = self.arr([range[0] for range in value_ranges])
            self._max = self.arr([range[1] for range in value_ranges])
            self._is_value_range = True
            # want an inclusive range on the min and max, so indexing with max should be valid
            self._resolution = (self._max - self._min) / (self.arr(self.shape) - 1)
        else:
            self._min = self.lib.zeros(self.dim, dtype=self.int)
            self._max = self.arr(source.shape, dtype=self.int) - 1
            self._is_value_range = False

        # flattened view of the source data
        self._d = source.reshape(-1)

    @classproperty
    @abc.abstractmethod
    def lib(cls):
        """Tensor/array library e.g. numpy and torch"""

    @property
    @abc.abstractmethod
    def int(self):
        """Return the integer type"""

    @property
    def raw_data(self):
        return self._d

    @abc.abstractmethod
    def arr(self, *args, **kwargs):
        """Create an array"""

    @abc.abstractmethod
    def zeros(self, *args, **kwargs):
        """Create a zero array"""

    @classmethod
    @abc.abstractmethod
    def cast(cls, arr, dtype):
        """Cast an arr into a dtype"""

    @classmethod
    @abc.abstractmethod
    def all(cls, arr, dim=0):
        """Logically evaluate to true iff all elements of arr is true"""

    @classmethod
    @abc.abstractmethod
    def is_valid_arr_value(cls, val, valid):
        """Evaluate if the value is an array of compatible size with valid mask"""

    @classmethod
    @abc.abstractmethod
    def ravel_multi_index(cls, key, shape):
        """Flatten keys from N x d to N"""

    @abc.abstractmethod
    def unravel_key(self, key):
        """Expand keys from N to N x d"""

    @classmethod
    @abc.abstractmethod
    def transpose(cls, arr):
        """Transpose dimensions 0 and 1"""

    @classmethod
    @abc.abstractmethod
    def repeat(cls, arr, repeats):
        """Repeat an array along its dimensions certain times"""

    @classmethod
    @abc.abstractmethod
    def cat(cls, arrs, dim=0):
        """Concatenate a sequence of arrays along a dimension"""

    @classmethod
    @abc.abstractmethod
    def stack(cls, arrs, dim=0):
        """Stack arrs along a newly created, specified dimension"""

    def is_key_ravelled(self, key):
        return len(key.shape) == 1 or (key.shape[-1] == 1 and self.dim != 1)

    def ensure_index_key(self, key, force=False):
        if self.is_key_ravelled(key):
            key = self.unravel_key(key.reshape(-1))
        # convert key from value ranges to indices if necessary
        if self._is_value_range and (force or key.dtype != self.int):
            index_key = self.stack(
                [self.cast(self.lib.round((key[..., i] - self._min[i]) / self._resolution[i]), self.int) for i in
                 range(self.dim)],
                dim=-1)
            key = index_key
        return key

    def ensure_value_key(self, key, force=False):
        if self.is_key_ravelled(key):
            key = self.unravel_key(key.reshape(-1))
        # convert key from indices to value ranges if necessary
        if self._is_value_range and (force or key.dtype != self.dtype):
            value_key = self.stack(
                [self.cast(key[..., i] * self._resolution[i] + self._min[i], self.dtype) for i in range(self.dim)],
                dim=-1)
            key = value_key
        return key

    def get_valid_values(self, key):
        return self.all(
            self.stack([(self._min[i] <= key[..., i]) & (key[..., i] <= self._max[i]) for i in range(self.dim)]),
            dim=0)

    def get_valid_ravel_indices(self, key):
        """
        Ravel a N x d key into a N length key of ravelled indices
        :param key: N x d key (could be values or indices depending on underlying data)
        :return: ravelled indices of length N along with boolean validity mask of length N
        """
        # check if shorthand is used where batch indices are not specified
        if key.shape[-1] == self.dim - 1:
            B = key.shape[0]
            batch_index = self.repeat(self.lib.arange(B).reshape(B, 1, 1), (1, key.shape[-2], 1))
            key = self.cat((batch_index, key), dim=-1)

        # flatten batch dimensions
        key = key.reshape(-1, key.shape[-1])

        # eliminate keys outside query
        if self.check_safety:
            valid = self.get_valid_values(key)
            key = key[valid]
        else:
            valid = True
        key = self.ensure_index_key(key)
        # flatten
        flat_key = self.ravel_multi_index(key, self.shape)
        return flat_key, valid

    def __getitem__(self, key):
        """
        Get a batch (size N) from the data with dimension d, indexed by a multidimensional key
        :param key: N x d query, with each row corresponding to one element to look up
        :return:
        """
        orig_key_shape = key.shape

        flat_key, valid = self.get_valid_ravel_indices(key)
        if self.check_safety:
            N = valid.shape[0]
            res = self.zeros(N, dtype=self.dtype)
            res[valid] = self._d[flat_key]
            if callable(self.invalid_value):
                invalid_entries = ~valid.reshape(key.shape[:-1])
                res[~valid] = self.invalid_value(key[invalid_entries])
            else:
                res[~valid] = self.invalid_value
        else:
            res = self._d[flat_key]

        return res.reshape(list(orig_key_shape[:-2]) + [-1])

    def __setitem__(self, key, value):
        """
        Batch (size N) assignment
        :param key: N x d query, with each row corresponding to one element to look up
        :param value: value of compatible type and shape
        :return:
        """
        flat_key, valid = self.get_valid_ravel_indices(key)
        if self.is_valid_arr_value(value, valid):
            value = value.reshape(-1)
            if self.check_safety:
                value = value[valid]
        self._d[flat_key] = value
